from abc import ABC, abstractmethod

from slugify import slugify

from kart.utils import KartDict, KartMap, paginate


class Mapper(ABC):
    """Base mapper class"""

    @abstractmethod
    def map(self, config: dict, site: KartDict) -> KartMap:
        """Takes the site as inputs and outputs a site map"""


class RuleMapper(Mapper):
    """Mapper that uses user defined functions to build the map"""

    def __init__(self, rules: list = []):
        """Initializes the mapper with a list of functions"""
        self.rules = rules  # a rule is a function

    def map(self, config: dict, site: KartDict) -> KartMap:
        urls = {}
        for rule in self.rules:
            urls.update(rule(site))
        return urls


class ManualMapper(Mapper):
    """Mapper that adds user defined pages to the map"""

    def __init__(self, pages: dict = {}):
        """Initializes the mapper with a list of pages"""
        self.pages = pages

    def map(self, config: dict, site: KartDict) -> KartMap:
        return self.pages


class DefaultCollectionMapper(Mapper):
    """Mapper intended to be used with DefaultCollectionMiner"""

    def __init__(
        self,
        collection: str,
        template: str = "item.html",
        base_url: str = "",
        renderer: str = "default_site_renderer",
    ):
        self.template = template
        self.base_url = base_url
        self.collection = collection
        self.renderer = renderer

    def map(self, config: dict, site: KartDict) -> KartMap:
        urls = {}
        collection = site[self.collection]
        for object in collection:
            slug = f"{self.collection}.{object['slug']}"
            url = f"/{slugify(self.collection)}/{slugify(object['slug'])}/"
            if "template" in object:
                template = object["template"]
            else:
                template = self.template
            page = {
                "url": self.base_url + url,
                "data": object,
                "template": template,
                "renderer": self.renderer,
            }
            urls[slug] = page
        return urls


class DefaultPageMapper(Mapper):
    """Mapper intended to be used with DefaultPageMiner"""

    def __init__(
        self, template: str = "page.html", renderer: str = "default_site_renderer"
    ):
        self.template = template
        self.renderer = renderer

    def map(self, config: dict, site: KartDict) -> KartMap:
        urls = {}
        for page in site["pages"]:
            slug = page["slug"]
            if "url" in page:
                url = page["url"]
            elif slug == "index":
                url = "/"
            else:
                url = f"/{slugify(slug)}/"
            if "template" in page:
                template = page["template"]
            else:
                template = self.template
            page = {
                "url": url,
                "data": page,
                "template": template,
                "renderer": self.renderer,
            }
            urls[slug] = page
        return urls


class DefaultIndexMapper(Mapper):
    """Mapper that creates the index pages of a collection"""

    def __init__(
        self,
        collection: str,
        template: str = "index.html",
        base_url: str = "",
        renderer: str = "default_site_renderer",
    ):
        self.template = template
        self.base_url = base_url
        self.collection = collection
        self.renderer = renderer

    def map(self, config: dict, site: KartDict) -> KartMap:
        items = list(site[self.collection])
        filtered_items = items[config["pagination"]["skip"] :]
        paginated_map = paginate(
            objects=filtered_items,
            per_page=config["pagination"]["per_page"],
            template=self.template,
            base_url=self.base_url + "/index/page/",
            slug=f"{self.collection}_index",
            renderer=self.renderer,
            additional_data={},
        )
        paginated_map[f"{self.collection}_index.1"]["url"] = self.base_url + "/"
        return paginated_map


class DefaultTaxonomyMapper(Mapper):
    """Mapper that creates the index pages of a taxonomy of a collection"""

    def __init__(
        self,
        collection: str,
        taxonomy: str,
        template: str = "tag.html",
        base_url: str = "",
        renderer: str = "default_site_renderer",
    ):
        self.template = template
        self.base_url = base_url
        self.collection = collection
        self.taxonomy = taxonomy
        self.renderer = renderer

    def map(self, config: dict, site: KartDict) -> KartMap:
        urls = {}
        for taxonomy in site[self.taxonomy]:
            slug = taxonomy["slug"]
            items = site[self.collection]
            filtered_items = []
            for item in items:
                if self.taxonomy not in item:
                    continue
                if isinstance(item[self.taxonomy], str):
                    if item[self.taxonomy] == slug:
                        filtered_items.append(item)
                elif isinstance(item[self.taxonomy], list):
                    if slug in item[self.taxonomy]:
                        filtered_items.append(item)
            urls.update(
                paginate(
                    objects=filtered_items,
                    per_page=config["pagination"]["per_page"],
                    template=self.template,
                    base_url=self.base_url + f"/{self.taxonomy}/{slug}/",
                    slug=f"{self.taxonomy}.{slug}",
                    renderer=self.renderer,
                    additional_data=taxonomy,
                )
            )
        return urls


class DefaultFeedMapper(Mapper):
    """Mapper that adds an entry for DefaultFeedMapper"""

    def __init__(self, collections: list = [], renderer: str = "default_feed_renderer"):
        self.collections = collections
        self.renderer = renderer

    def map(self, config: dict, site: KartDict) -> KartMap:
        return {
            "feed": {
                "url": "/atom.xml",
                "data": {"collections": self.collections},
                "template": "",
                "renderer": self.renderer,
            }
        }


class DefaultSitemapMapper(Mapper):
    """Mapper that adds an entry for DefaultFeedMapper"""

    def map(elf, config: dict, site: KartDict) -> KartMap:
        return {
            "sitemap": {
                "url": "/sitemap.xml",
                "data": {},
                "template": "",
                "renderer": "default_sitemap_renderer",
            },
        }


class DefaultStaticFilesMapper(Mapper):
    """Mapper that adds an entry for DefaultFeedMapper"""

    def map(self, config: dict, site: KartDict) -> KartMap:
        return {
            "static": {
                "url": "/static/*",
                "data": {},
                "template": "",
                "renderer": "default_static_files_renderer",
            }
        }


class DefaultRootDirMapper(Mapper):
    """Mapper that adds an entry for DefaultFeedMapper"""

    def map(self, config: dict, site: KartDict) -> KartMap:
        return {
            "root": {
                "url": "/*",
                "data": {},
                "template": "",
                "renderer": "default_root_dir_renderer",
            }
        }


class DefaultRobotsTxtMapper(Mapper):
    """Mapper that adds an entry for DefaultRobotsTxtRenderer"""

    def __init__(self, renderer: str = "default_robotstxt_renderer"):
        self.renderer = renderer

    def map(self, config: dict, site: KartDict) -> KartMap:
        if "robots" in site:
            return {
                "robots": {
                    "url": "/robots.txt",
                    "content": site["robots"]["content"],
                    "renderer": self.renderer,
                }
            }



from abc import ABC, abstractmethod

from kart.utils import KartDict, KartMap


class ContentModifier(ABC):
    """Base ContentModifier class"""

    @abstractmethod
    def modify(self, config: dict, site: KartDict):
        """Takes as input the site and modifies it"""


class MapModifier(ABC):
    """Base MapModifier class"""

    @abstractmethod
    def modify(self, config: dict, site: KartDict, map: KartMap):
        """Takes as input the site and the site map and modifies them"""


class RuleContentModifier(ContentModifier):
    """ContentModifier that executes a list of user defined functions"""

    def __init__(self, rules: list = []):
        self.rules = rules  # a rule is a function

    def modify(self, config: dict, site: KartDict):
        for rule in self.rules:
            rule(site)


class RuleMapModifier(MapModifier):
    """MapModifier that executes a list of user defined functions"""

    def __init__(self, rules: list = []):
        self.rules = rules  # a rule is a function

    def modify(self, config: dict, site: KartDict, map: KartMap):
        for rule in self.rules:
            rule(map, site)


class CollectionSorter(ContentModifier):
    """Modifier which sorts a collection based on a key"""

    def __init__(self, collection, key, reverse=False):
        self.collection = collection
        self.key = key
        self.reverse = reverse

    def modify(self, config: dict, site: KartDict):
        """Sorts the collection"""
        data = site[self.collection]
        sorted_data = sorted(data.items(), key=lambda x: x[1][self.key])
        if self.reverse:
            sorted_data.reverse()
        site[self.collection] = KartDict(sorted_data)



from abc import ABC, abstractmethod
from datetime import datetime, time
from shutil import copytree
from http.server import SimpleHTTPRequestHandler
from multiprocessing import Pool
from pathlib import Path

from dateutil import tz
from jinja2 import Environment, FileSystemLoader

from kart.ext.markdown import markdown_to_html, markdown_to_toc
from kart.utils import KartDict, KartMap, date_to_string


class Renderer(ABC):
    """Base Renderer class"""

    @abstractmethod
    def render(self, config: dict, site: KartDict, map: KartMap, build_location: str):
        """Renders each page indented for this renderer"""

    def start_serving(self, config: dict):
        """Start the dev server, if necessary"""

    @abstractmethod
    def serve(
        self, http_handler, page: dict, config: dict, site: KartDict, map: KartMap
    ):
        """Serves the the requested page"""

    def stop_serving(self, config: dict):
        """Stops the dev server, if necessary"""


class DefaultFileRenderer(Renderer):
    """Base class for renderers that render file individually"""

    @abstractmethod
    def __init__(self, name: str):
        """Initializes renderer. Must set the ``name`` and ``content_type`` variables"""

    @abstractmethod
    def render_single(
        self, page: dict, config: dict, site: KartDict, map: KartMap
    ) -> str:
        """Renders a single file"""

    def render(self, config, site, map, build_location):
        for page in map:
            if page["renderer"] != self.name:
                continue
            rendered_file = self.render_single(page, config, site, map)
            path = Path(build_location) / Path(*Path(page["url"]).parts[1:])
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("w") as f:
                f.write(rendered_file)

    def serve(
        self, http_handler, page: dict, config: dict, site: KartDict, map: KartMap
    ):
        http_handler.send_response(200)
        http_handler.send_header("Content-type", self.content_type)
        http_handler.end_headers()
        content = self.render_single(page, config, site, map)
        http_handler.wfile.write(bytes(content, "utf-8"))


class DefaultDirectoryRenderer(Renderer):
    """Base class for renderers that render directories"""

    @abstractmethod
    def __init__(self, name: str, directory: str):
        """Initializes renderer. Must set the ``name``, ``dir`` and ``base_url`` variables"""

    def render(self, config: dict, site: KartDict, map: KartMap, build_location: str):
        """Copies the entire directory to the target destination"""
        for page in map:
            if page["renderer"] != self.name:
                continue
            destination = Path(build_location) / Path(*Path(page["url"][:-1]).parts[1:])
            copytree(Path(self.dir), destination, dirs_exist_ok=True)

    def serve(
        self, http_handler, page: dict, config: dict, site: KartDict, map: KartMap
    ):
        http_handler.path = self.base_url + http_handler.path
        return SimpleHTTPRequestHandler.do_GET(http_handler)


class DefaultSiteRenderer(DefaultFileRenderer):
    """Default renderer for rendering html files"""

    def __init__(
        self,
        name: str = "default_site_renderer",
        template_folder: str = "templates",
        filters: dict = {
            "html": markdown_to_html,
            "toc": markdown_to_toc,
            "date_to_string": date_to_string,
        },
        process_count: int = 1,
    ):
        self.name = name
        self.content_type = "text/html"
        self.template_folder = template_folder
        self.process_count = process_count
        self.env = Environment(loader=FileSystemLoader(self.template_folder))
        self.env.filters.update(filters)

    def render_single(
        self, page: dict, config: dict, site: KartDict, map: KartMap
    ) -> str:
        template = self.env.get_template(page["template"])
        page = {**page["data"], "url": page["url"]}
        return template.render(page=page, config=config, site=site, url=map.url)

    def _render_single(
        self, key: str, config: dict, site: KartDict, map: KartMap, build_location: str
    ):
        """Renderers a single file and saves it to the disk"""
        page = map[key]
        if page["renderer"] != self.name:
            return
        rendered_file = self.render_single(page, config, site, map)
        path = Path(build_location) / Path(*Path(page["url"]).parts[1:])
        path = path / "index.html"
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            f.write(rendered_file)

    def render(
        self, config: dict, site: KartDict, map: KartMap, build_location: str = "_site"
    ):
        """Renderers all the files with a multiprocessing Pool for faster build times"""
        if self.process_count == 1:
            for key in map.keys():
                self._render_single(key, config, site, map, build_location)
        else:
            with Pool(self.process_count) as p:
                data = ((key, config, site, map, build_location) for key in map.keys())
                p.starmap(self._render_single, data)


class DefaultFeedRenderer(DefaultFileRenderer):
    """Renders an atom feed file"""

    def __init__(self, name: str = "default_feed_renderer"):
        """Initializes renderer. Sets the ``name`` and ``content_type`` variables"""
        self.name = name
        self.content_type = "application/xml"

    def render_single(
        self, page: dict, config: dict, site: KartDict, map: KartMap
    ) -> str:
        """Creates the atom feeds"""
        feed_entries = []
        for collection in page["data"]["collections"]:
            for object in site[collection]:
                feed_entries.append([map.url(collection, object["slug"]), object])
        feed_entries.sort(key=lambda x: x[1]["date"], reverse=True)
        atom = '<feed xmlns="http://www.w3.org/2005/Atom">'
        atom += f'<id>"{map.url("/")}"</id>'
        atom += f"<title>{config['name']}</title>"
        timezone = config["timezone"]
        updated_time = datetime.now().replace(tzinfo=tz.gettz(timezone))
        atom += f"<updated>{updated_time.isoformat()}</updated>"
        atom += f'<link href="{map.url("/")}"/>'
        atom += f'<link href="{map.url(page["url"])}" rel="self"/>'
        for url, entry in feed_entries:
            atom += "<entry>"
            atom += f"<id>{url}</id>"
            title = entry["title"] if "title" in entry.keys() else entry["name"]
            atom += f"<title>{title}</title>"
            entry_time = datetime.combine(entry["date"], time(12))
            entry_time.replace(tzinfo=tz.gettz(timezone))
            atom += f"<updated>{entry_time.isoformat()}</updated>"
            if "description" in entry.keys():
                atom += f'<summary>{entry["description"]}</summary>'
            atom += f'<link href="{url}" link="alternate"/>'
            atom += "</entry>"
        atom += "</feed>"
        return atom


class DefaultSitemapRenderer(DefaultFileRenderer):
    """Renders the sitemap of the site"""

    def __init__(self, name: str = "default_sitemap_renderer"):
        """Initializes renderer. Sets the ``name`` and ``content_type`` variables"""
        self.name = name
        self.content_type = "application/xml"

    def render_single(self, page: dict, config: dict, site: KartDict, map: KartMap):
        """Uses the site ``map`` variable to create the sitemap"""
        sitemap = '<?xml version="1.0" encoding="UTF-8"?>'
        sitemap += '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">'
        for x in map:
            if x["renderer"] != "default_site_renderer":
                continue
            sitemap += f"<url><loc>{map.url(x['url'])}</loc></url>"
        sitemap += "</urlset>"
        return sitemap


class DefaultStaticFilesRenderer(DefaultDirectoryRenderer):
    """Renders all files in the ``static`` directory"""

    def __init__(
        self, name: str = "default_static_files_renderer", directory: str = "static"
    ):
        """Initializes renderer. Sets the ``name``, ``dir`` and ``content_type`` variables"""
        self.name = name
        self.dir = directory
        self.base_url = ""


class DefaultRootDirRenderer(DefaultDirectoryRenderer):
    """Renders all files in the ``root`` directory"""

    def __init__(
        self, name: str = "default_root_dir_renderer", directory: str = "root"
    ):
        """Initializes renderer. Sets the ``name``, ``dir`` and ``content_type`` variables"""
        self.name = name
        self.dir = directory
        self.base_url = "/root"


class DefaultRobotsTxtRenderer(DefaultFileRenderer):
    """Renders robots.txt"""

    def __init__(self, name: str = "default_robotstxt_renderer"):
        self.name = name
        self.content_type = "text/plain"

    def render_single(
        self, page: dict, config: dict, site: KartDict, map: KartMap
    ) -> str:
        template = Environment().from_string(page["content"])
        robots = template.render(config=config, site=site, url=map.url)
        return robots



import xml.etree.ElementTree as xml
from datetime import date
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Any, Dict

from jinja2 import pass_context
from jinja2.runtime import Context

from kart.miners import DefaultMiner
from kart.utils import id_from_path, str_to_bool


@pass_context
def asciidoc_to_html(context: Context, asciidoc: str) -> str:
    """Converts asciidoc data to html"""
    parsed_asciidoc = context.environment.from_string(asciidoc).render(context)
    config = context["config"]["code_highlighting"]
    pygments_header = ":source-highlighter: pygments\n"
    if config["noclasses"]:
        pygments_header += ":pygments-css: style\n"
        pygments_header += f':pygments-style: {config["style"]}\n'
    parsed_asciidoc = pygments_header + parsed_asciidoc
    p = Popen(["asciidoctor", "-e", "-"], stdin=PIPE, stdout=PIPE)
    return p.communicate(input=parsed_asciidoc.encode("utf-8"))[0].decode("utf-8")


def asciidoc_to_toc(asciidoc: str) -> str:
    """Extracts a list of header from asciidoc data"""
    asciidoc = ":toc:\n:toclevels: 4\n" + asciidoc
    p = Popen(["asciidoctor", "-e", "-a", "toc", "-"], stdin=PIPE, stdout=PIPE)
    html = p.communicate(input=asciidoc.encode("utf-8"))[0].decode("utf-8")
    toc_xml = xml.fromstring("</div>".join(html.split("</div>")[:2]) + "</div>")[1]

    def parse(ul):
        toc = []
        for li in ul:
            for x in li:
                if x.tag == "a":
                    toc.append(
                        {
                            "title": x.text,
                            "id": x.attrib["href"][1:],
                            "level": int(ul.attrib["class"][-1:]) + 1,
                        }
                    )
                elif x.tag == "ul":
                    toc.extend(parse(x))
        return toc

    return parse(toc_xml)


def parse_asciidoc_header(header: str) -> Dict[str, Any]:
    lines = header.splitlines()[1:]
    metadata = {}
    metadata["title"] = lines[0][2:]
    lines.pop(0)
    if not lines[0].startswith("//") and not lines[0].startswith(":"):
        metadata["author"] = lines[0]
        lines.pop(0)
    for line in lines:
        if line.startswith("//"):
            continue
        key = line.split(":")[1]
        value = ":".join(line.split(":")[2:]).strip()
        if key == "tags":
            value = [x.strip() for x in value.split(",")]
        if key == "date":
            value = date.fromisoformat(value)
        if value == "":
            # https://docs.asciidoctor.org/asciidoc/latest/attributes/boolean-attributes/
            value = True
        try:
            value = str_to_bool(value)
        except:
            pass
        metadata[key] = value
    return metadata


class AsciidocMiner(DefaultMiner):
    """Base miner that implements collect_single_file() for markup files"""

    extensions = (".adoc",)

    def collect_single_file(self, file: Path, config: dict) -> str:
        """
        Stores the data included in the frontmatter in a dictionary, adds
        the content in the ``content`` field and then return the dictionary
        """
        with file.open("r") as f:
            data = f.read().split("\n\n")
            metadata = parse_asciidoc_header(data[0])
            if "draft" in metadata and metadata["draft"] and not config["draft_mode"]:
                return
            content = "\n\n".join(data[1:])
            slug = id_from_path(self.dir, file)
            data = {}
            data["markup"] = "asciidoc"
            data.update(metadata)
            data["content"] = content
            data["slug"] = slug
            return {slug: data}


class AsciidocCollectionMiner(AsciidocMiner):
    """Miner that looks for data in the ``collections`` folder"""

    def __init__(self, collection: str, directory: str = "collections"):
        "Initializes miner. Sets the ``name``, ``dir`` and ``collection`` variables"
        self.collection = collection
        self.dir = Path() / directory / collection
        self.name = self.collection


class AsciidocTaxonomyMiner(AsciidocMiner):
    """Miner that looks for data in the ``taxonomy`` folder"""

    def __init__(self, taxonomy: str, directory: str = "taxonomies"):
        "Initializes miner. Sets the ``name``, ``dir`` and ``taxonomy`` variables"
        self.taxonomy = taxonomy
        self.dir = Path() / directory / taxonomy
        self.name = self.taxonomy


class AsciidocPageMiner(AsciidocMiner):
    """Miner that get data from the ``pages`` folder"""

    def __init__(self, directory: str = "pages"):
        "Initializes miner. Sets the ``name`` and ``dir`` variables"
        self.dir = Path(directory)
        self.name = "pages"



from pathlib import Path

import tomli

from kart.miners import DefaultMiner
from kart.utils import id_from_path


class TomlDataMiner(DefaultMiner):
    """Miner that get yaml data from the ``data`` folder"""

    extensions = (".toml",)

    def __init__(self, directory: str = "data"):
        "Initializes miner. Sets the ``name`` and ``dir`` variables"
        self.dir = Path(directory)
        self.name = "data"

    def collect_single_file(self, file: Path, config: dict) -> str:
        """Users tomli to get data from a toml file"""
        with file.open("rb") as f:
            slug = id_from_path(self.dir, file)
            return {slug: tomli.load(f)}
